@@ -1,14 +1,15 @@
 # -*- coding: utf-8 -*-
 import logging
 import time
+import uuid
 from typing import List
 
 import pandas as pd
 from sqlalchemy.orm import Session
 
-from zvdata import IntervalLevel, Mixin
+from zvdata import IntervalLevel, Mixin, EntityMixin
 from zvdata.api import get_entities, get_data
-from zvdata.contract import get_db_session
+from zvdata.contract import get_db_session, get_schema_columns
 from zvdata.utils.time_utils import to_pd_timestamp, TIME_FORMAT_DAY, to_time_str, \
     evaluate_size_from_timestamp, is_in_same_interval
 from zvdata.utils.utils import fill_domain_from_dict
@@ -18,10 +19,10 @@ class Meta(type):
     def __new__(meta, name, bases, class_dict):
         cls = type.__new__(meta, name, bases, class_dict)
         # register the recorder class to the data_schema
-        if hasattr(cls, 'data_schema'):
+        if hasattr(cls, 'data_schema') and hasattr(cls, 'provider'):
             if cls.data_schema and issubclass(cls.data_schema, Mixin):
-                print(f'{cls.data_schema.__name__}')
-                cls.data_schema.register_recorder_cls(cls)
+                print(f'{cls.__name__}:{cls.data_schema.__name__}')
+                cls.data_schema.register_recorder_cls(cls.provider, cls)
         return cls
 
 
@@ -52,6 +53,7 @@ class Recorder(metaclass=Meta):
 
         assert self.provider is not None
         assert self.data_schema is not None
+        assert self.provider in self.data_schema.providers
 
         self.batch_size = batch_size
         self.force_update = force_update
@@ -65,13 +67,15 @@ class Recorder(metaclass=Meta):
         raise NotImplementedError
 
     def sleep(self):
-        time.sleep(self.sleeping_time)
+        if self.sleeping_time > 0:
+            self.logger.info(f'sleeping {self.sleeping_time} seconds')
+            time.sleep(self.sleeping_time)
 
 
 class RecorderForEntities(Recorder):
     # overwrite them to fetch the entity list
     entity_provider: str = None
-    entity_schema = None
+    entity_schema: EntityMixin = None
 
     def __init__(self,
                  entity_type='stock',
@@ -116,6 +120,10 @@ class RecorderForEntities(Recorder):
         self.init_entities()
 
     def init_entities(self):
+        """
+        init the entities which we would record data for
+
+        """
         if self.entity_provider == self.provider and self.entity_schema == self.data_schema:
             self.entity_session = self.session
         else:
@@ -254,20 +262,24 @@ class TimeSeriesDataRecorder(RecorderForEntities):
         :param original_data:
         """
 
-        updated = False
+        got_new_data = False
 
         # if the domain is directly generated in record method, we just return it
         if isinstance(original_data, self.data_schema):
-            return updated, original_data
+            got_new_data = True
+            return got_new_data, original_data
 
         the_id = self.generate_domain_id(entity, original_data)
 
+        # optional way
+        # item = self.session.query(self.data_schema).get(the_id)
+
         items = get_data(data_schema=self.data_schema, session=self.session, provider=self.provider,
                          entity_id=entity.id, filters=[self.data_schema.id == the_id], return_type='domain')
 
         if items and not self.force_update:
             self.logger.info('ignore the data {}:{} saved before'.format(self.data_schema, the_id))
-            return updated, None
+            return got_new_data, None
 
         if not items:
             timestamp_str = original_data[self.get_original_time_field()]
@@ -277,16 +289,23 @@ class TimeSeriesDataRecorder(RecorderForEntities):
             except Exception as e:
                 self.logger.exception(e)
 
-            domain_item = self.data_schema(id=the_id,
-                                           code=entity.code,
-                                           entity_id=entity.id,
-                                           timestamp=timestamp)
+            if 'name' in get_schema_columns(self.data_schema):
+                domain_item = self.data_schema(id=the_id,
+                                               code=entity.code,
+                                               name=entity.name,
+                                               entity_id=entity.id,
+                                               timestamp=timestamp)
+            else:
+                domain_item = self.data_schema(id=the_id,
+                                               code=entity.code,
+                                               entity_id=entity.id,
+                                               timestamp=timestamp)
+            got_new_data = True
         else:
             domain_item = items[0]
-            updated = True
 
         fill_domain_from_dict(domain_item, original_data, self.get_data_map())
-        return updated, domain_item
+        return got_new_data, domain_item
 
     def persist(self, entity, domain_list):
         """
@@ -315,7 +334,14 @@ class TimeSeriesDataRecorder(RecorderForEntities):
             self.session.commit()
 
     def on_finish(self):
-        self.session.close()
+        try:
+            if self.session:
+                self.session.close()
+
+            if self.entity_session:
+                self.entity_session.close()
+        except Exception as e:
+            self.logger.error(e)
 
     def on_finish_entity(self, entity):
         pass
@@ -325,8 +351,11 @@ class TimeSeriesDataRecorder(RecorderForEntities):
         unfinished_items = self.entities
         raising_exception = None
         while True:
-            for entity_item in unfinished_items:
+            count = len(unfinished_items)
+            for index, entity_item in enumerate(unfinished_items):
                 try:
+                    self.logger.info(f'run to {index + 1}/{count}')
+
                     start_timestamp, end_timestamp, size, timestamps = self.evaluate_start_end_size_timestamps(
                         entity_item)
                     size = int(size)
@@ -358,6 +387,10 @@ class TimeSeriesDataRecorder(RecorderForEntities):
                         self.on_finish_entity(entity_item)
                         continue
 
+                    # sleep for a while to next entity
+                    if index != 0:
+                        self.sleep()
+
                     original_list = self.record(entity_item, start=start_timestamp, end=end_timestamp, size=size,
                                                 timestamps=timestamps)
 
@@ -365,11 +398,10 @@ class TimeSeriesDataRecorder(RecorderForEntities):
 
                     if original_list:
                         domain_list = []
-                        duplicate_count = 0
                         for original_item in original_list:
-                            updated, domain_item = self.generate_domain(entity_item, original_item)
+                            got_new_data, domain_item = self.generate_domain(entity_item, original_item)
 
-                            if not updated:
+                            if got_new_data:
                                 all_duplicated = False
 
                             # handle the case  generate_domain_id generate duplicate id
@@ -378,10 +410,10 @@ class TimeSeriesDataRecorder(RecorderForEntities):
                                 if duplicate:
                                     # regenerate the id
                                     if self.fix_duplicate_way == 'add':
-                                        duplicate_count += 1
-                                        domain_item.id = "{}_{}".format(domain_item.id, duplicate_count)
+                                        domain_item.id = "{}_{}".format(domain_item.id, uuid.uuid1())
                                     # ignore
                                     else:
+                                        self.logger.info(f'ignore original duplicate item:{domain_item.id}')
                                         continue
 
                                 domain_list.append(domain_item)
@@ -389,7 +421,7 @@ class TimeSeriesDataRecorder(RecorderForEntities):
                         if domain_list:
                             self.persist(entity_item, domain_list)
                         else:
-                            self.logger.info('just get {} duplicated data in this cycle'.format(len(original_list)))
+                            self.logger.info('just got {} duplicated data in this cycle'.format(len(original_list)))
 
                     # could not get more data
                     entity_finished = False
@@ -426,7 +458,6 @@ class TimeSeriesDataRecorder(RecorderForEntities):
                         self.on_finish_entity(entity_item)
                         continue
 
-                    time.sleep(self.sleeping_time)
                 except Exception as e:
                     self.logger.exception(
                         "recording data for entity_id:{},{},error:{}".format(entity_item.id, self.data_schema, e))
@@ -476,6 +507,7 @@ class FixedCycleDataRecorder(TimeSeriesDataRecorder):
     def get_latest_saved_record(self, entity):
         order = eval('self.data_schema.{}.desc()'.format(self.get_evaluated_time_field()))
 
+        # 对于k线这种数据，最后一个记录有可能是没完成的，所以取两个，总是删掉最后一个数据，更新之
         records = get_data(entity_id=entity.id,
                            provider=self.provider,
                            data_schema=self.data_schema,
@@ -483,7 +515,7 @@ class FixedCycleDataRecorder(TimeSeriesDataRecorder):
                            limit=2,
                            return_type='domain',
                            session=self.session,
-                           level=self.level.value)
+                           level=self.level)
         if records:
             # delete unfinished kdata
             if len(records) == 2:
@@ -528,9 +560,12 @@ class TimestampsDataRecorder(TimeSeriesDataRecorder):
                  real_time=False,
                  fix_duplicate_way='add',
                  start_timestamp=None,
-                 end_timestamp=None) -> None:
+                 end_timestamp=None,
+                 close_hour=0,
+                 close_minute=0) -> None:
         super().__init__(entity_type, exchanges, entity_ids, codes, batch_size, force_update, sleeping_time,
-                         default_size, real_time, fix_duplicate_way, start_timestamp, end_timestamp)
+                         default_size, real_time, fix_duplicate_way, start_timestamp, end_timestamp,
+                         close_hour=close_hour, close_minute=close_minute)
         self.security_timestamps_map = {}
 
     def init_timestamps(self, entity_item) -> List[pd.Timestamp]:
